@@ -1,9 +1,7 @@
 use crate::{
-    operators::BinOp,
-    util::{apply_unary_ops, CompositionOfUnaryOps},
+    operators::{BinOp, UnaryOp},
     ExParseError,
 };
-use itertools::Itertools;
 use smallvec::{smallvec, SmallVec};
 use std::fmt::Debug;
 
@@ -21,21 +19,10 @@ pub type FlatOpVec<T> = SmallVec<[FlatOp<T>; N_NODES_ON_STACK]>;
 /// will be executed after the binary operation in case of its existence.
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub struct FlatOp<T: Copy> {
-    unary_op: Option<CompositionOfUnaryOps<T>>,
+    unary_op: UnaryOp<T>,
     bin_op: BinOp<T>,
 }
 
-/// Nodes are inputs for binary operators. A node can be an expression, a number, or
-/// a variable.
-#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
-pub enum Node<T: Copy> {
-    Expr(Expression<T>),
-    Num(T),
-    /// The contained integer points to the index of the variable in the slice of
-    /// variables passed to [`eval`](Expression::eval).
-    Var(usize),
-}
-
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub enum FlatNodeKind<T: Copy> {
     Num(T),
@@ -45,21 +32,21 @@ pub enum FlatNodeKind<T: Copy> {
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub struct FlatNode<T: Copy> {
     kind: FlatNodeKind<T>,
-    unary_op: Option<CompositionOfUnaryOps<T>>,
+    unary_op: UnaryOp<T>,
 }
 
 impl<T: Copy> FlatNode<T> {
     pub fn from_kind(kind: FlatNodeKind<T>) -> FlatNode<T> {
         return FlatNode {
             kind: kind,
-            unary_op: None,
+            unary_op: UnaryOp::new(),
         };
     }
 }
 
 /// This is the core data type representing a flattened expression and the result of
-/// parsing a string. The expression is flattened to make efficient evaluation possible.
-/// Simplified, it consists of a [`SmallVec`](SmallVec) of nodes and a
+/// parsing a string. We use flattened expressions to make efficient evaluation possible.
+/// Simplified, a flat expression consists of a [`SmallVec`](SmallVec) of nodes and a
 /// [`SmallVec`](SmallVec) of operators that are applied to the nodes in an order following
 /// operator priorities.
 ///
@@ -82,7 +69,7 @@ impl<T: Copy> FlatNode<T> {
 /// The second argument `&[2.0, 1.5]` in the call of [`eval`](FlatEx::eval) specifies the
 /// variable values in the order of their occurrence in the string.
 /// In this example, we want to evaluate the expression for the varibale values `y=2.0` and `x=1.5`.
-/// Variables in the string to-be-parsed are all substrings that are no numbers, no 
+/// Variables in the string to-be-parsed are all substrings that are no numbers, no
 /// operators, and no parentheses.
 ///
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
@@ -93,14 +80,7 @@ pub struct FlatEx<T: Copy> {
     n_unique_vars: usize,
 }
 
-fn apply_uop_if_some<T: Copy>(uop: &Option<CompositionOfUnaryOps<T>>, val: T) -> T {
-    match uop {
-        None => val,
-        Some(uops) => apply_unary_ops(&uops, val),
-    }
-}
-
-impl<T: Copy> FlatEx<T> {
+impl<T: Copy + Debug> FlatEx<T> {
     /// Evaluates an expression with the given variable values and returns the computed
     /// result.
     ///
@@ -129,16 +109,12 @@ impl<T: Copy> FlatEx<T> {
             .nodes
             .iter()
             .map(|node| {
-                apply_uop_if_some(
-                    &node.unary_op,
-                    match node.kind {
-                        FlatNodeKind::Num(n) => n,
-                        FlatNodeKind::Var(idx) => vars[idx],
-                    },
-                )
+                node.unary_op.apply(match node.kind {
+                    FlatNodeKind::Num(n) => n,
+                    FlatNodeKind::Var(idx) => vars[idx],
+                })
             })
             .collect::<SmallVec<[T; 32]>>();
-        
         let mut ignore: SmallVec<[bool; N_NODES_ON_STACK]> = smallvec![false; N_NODES_ON_STACK];
         for (i, &bin_op_idx) in self.prio_indices.iter().enumerate() {
             let num_idx = self.prio_indices[i];
@@ -153,92 +129,51 @@ impl<T: Copy> FlatEx<T> {
             let num_1 = numbers[num_idx - shift_left];
             let num_2 = numbers[num_idx + shift_right];
             numbers[num_idx - shift_left] = {
-                let bop_res = (self.ops[bin_op_idx].bin_op.op)(num_1, num_2);
-                apply_uop_if_some(&self.ops[bin_op_idx].unary_op, bop_res)
+                let bop_res = (self.ops[bin_op_idx].bin_op.apply)(num_1, num_2);
+                self.ops[bin_op_idx].unary_op.apply(bop_res)
             };
             ignore[num_idx + shift_right] = true;
         }
         Ok(numbers[0])
     }
-
-    fn compile(&mut self) {
-        let mut num_inds = self.prio_indices.clone();
-        let mut used_prio_indices = ExprIdxVec::new();
-        for (i, &bin_op_idx) in self.prio_indices.iter().enumerate() {
-            let num_idx = num_inds[i];
-            let node_1 = &self.nodes[num_idx];
-            let node_2 = &self.nodes[num_idx + 1];
-            if let (FlatNodeKind::Num(num_1), FlatNodeKind::Num(num_2)) =
-                (&node_1.kind, &node_2.kind)
-            {
-                let num_1 = apply_uop_if_some(&node_1.unary_op, *num_1);
-                let num_2 = apply_uop_if_some(&node_2.unary_op, *num_2);
-                let val = apply_uop_if_some(
-                    &self.ops[bin_op_idx].unary_op,
-                    (self.ops[bin_op_idx].bin_op.op)(num_1, num_2),
-                );
-                self.nodes[num_idx] = FlatNode {
-                    kind: FlatNodeKind::Num(val),
-                    unary_op: None,
-                };
-                self.nodes.remove(num_idx + 1);
-                // reduce indices after removed position
-                for num_idx_after in num_inds.iter_mut() {
-                    if *num_idx_after > num_idx {
-                        *num_idx_after = *num_idx_after - 1;
-                    }
-                }
-                used_prio_indices.push(bin_op_idx);
-            } else {
-                break;
-            }
-        }
-
-        self.ops = self
-            .ops
-            .iter()
-            .enumerate()
-            .filter(|(i, _)| !used_prio_indices.contains(i))
-            .map(|x| x.1.clone())
-            .collect();
-
-        self.prio_indices = prioritized_indices_flat(&self.ops, &self.nodes);
-    }
 }
 
-fn flatten_vecs<T: Copy>(expr: &Expression<T>, prio_offset: i32) -> (FlatNodeVec<T>, FlatOpVec<T>) {
+fn flatten_vecs<T: Copy>(
+    deep_expr: &DeepEx<T>,
+    prio_offset: i32,
+) -> (FlatNodeVec<T>, FlatOpVec<T>) {
     let mut flat_nodes = FlatNodeVec::<T>::new();
     let mut flat_ops = FlatOpVec::<T>::new();
 
-    for (node_idx, node) in expr.nodes.iter().enumerate() {
+    for (node_idx, node) in deep_expr.nodes.iter().enumerate() {
         match node {
-            Node::Num(num) => {
+            DeepNode::Num(num) => {
                 let flat_node = FlatNode::from_kind(FlatNodeKind::Num(*num));
                 flat_nodes.push(flat_node);
             }
-            Node::Var(idx) => {
+            DeepNode::Var(idx) => {
                 let flat_node = FlatNode::from_kind(FlatNodeKind::Var(*idx));
                 flat_nodes.push(flat_node);
             }
-            Node::Expr(e) => {
+            DeepNode::Expr(e) => {
                 let (mut sub_nodes, mut sub_ops) = flatten_vecs(e, prio_offset + 100i32);
                 flat_nodes.append(&mut sub_nodes);
                 flat_ops.append(&mut sub_ops);
             }
         };
-        if node_idx < expr.bin_ops.len() {
+        if node_idx < deep_expr.bin_ops.len() {
             let prio_adapted_bin_op = BinOp {
-                op: expr.bin_ops[node_idx].op,
-                prio: expr.bin_ops[node_idx].prio + prio_offset,
+                apply: deep_expr.bin_ops[node_idx].apply,
+                prio: deep_expr.bin_ops[node_idx].prio + prio_offset,
             };
             flat_ops.push(FlatOp {
                 bin_op: prio_adapted_bin_op,
-                unary_op: None,
+                unary_op: UnaryOp::new(),
             });
         }
     }
 
-    if expr.unary_ops.len() > 0 {
+    if deep_expr.unary_op.len() > 0 {
         if flat_ops.len() > 0 {
             // find the last binary operator with the lowest priority of this expression,
             // since this will be executed as the last one
@@ -246,86 +181,97 @@ fn flatten_vecs<T: Copy>(expr: &Expression<T>, prio_offset: i32) -> (FlatNodeVec
                 None => panic!("cannot have more than one flat node but no binary ops"),
                 Some(x) => x,
             };
-            let mut new_uops = expr.unary_ops.clone();
-            let mut existing_uops = match low_prio_op.unary_op.clone() {
-                Some(uops) => uops,
-                None => CompositionOfUnaryOps::<T>::new(),
-            };
-            new_uops.append(&mut existing_uops);
-            *low_prio_op = FlatOp {
-                bin_op: low_prio_op.bin_op,
-                unary_op: Some(new_uops),
-            }
+            low_prio_op
+                .unary_op
+                .append_front(&mut deep_expr.unary_op.clone());
         } else {
-            let mut new_op = expr.unary_ops.clone();
-            flat_nodes[0].unary_op = match flat_nodes[0].unary_op.clone() {
-                None => Some(new_op),
-                Some(mut uops) => {
-                    new_op.append(&mut uops);
-                    Some(new_op)
-                }
-            }
+            flat_nodes[0]
+                .unary_op
+                .append_front(&mut deep_expr.unary_op.clone());
         }
     }
     (flat_nodes, flat_ops)
 }
 
+/// A deep node can be an expression, a number, or
+/// a variable.
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
-pub struct Expression<T: Copy> {
+pub enum DeepNode<T: Copy> {
+    Expr(DeepEx<T>),
+    Num(T),
+    /// The contained integer points to the index of the variable in the slice of
+    /// variables passed to [`eval`](Expression::eval).
+    Var(usize),
+}
+
+/// A deep expression evaluates co-recursively since its nodes can contain other deep
+/// expressions.
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
+pub struct DeepEx<T: Copy> {
     /// Nodes can be numbers, variables, or other expressions.
-    nodes: Vec<Node<T>>,
+    nodes: Vec<DeepNode<T>>,
     /// Binary operators applied to the nodes according to their priority.
     bin_ops: BinOpVec<T>,
     /// Unary operators are applied to the result of evaluating all nodes with all
-    /// binary operators. The last unary operator is applied first to the result
-    /// of the evaluation of nodes and binary operators
-    unary_ops: CompositionOfUnaryOps<T>,
+    /// binary operators.
+    unary_op: UnaryOp<T>,
     prio_indices: ExprIdxVec,
 }
 
 fn prioritized_indices_flat<T: Copy>(ops: &[FlatOp<T>], nodes: &FlatNodeVec<T>) -> ExprIdxVec {
-    let mut indices: ExprIdxVec = (0..ops.len()).collect();
-    indices.sort_by(|i1, i2| {
-        let (prio_i1, prio_i2) = match (&nodes[*i1].kind, &nodes[*i2].kind) {
+    let prio_increase =
+        |bin_op_idx: usize| match (&nodes[bin_op_idx].kind, &nodes[bin_op_idx + 1].kind) {
             (FlatNodeKind::Num(_), FlatNodeKind::Num(_)) => {
                 let prio_inc = 5;
-                (
-                    &ops[*i1].bin_op.prio * 10 + prio_inc,
-                    &ops[*i2].bin_op.prio * 10 + prio_inc,
-                )
+                &ops[bin_op_idx].bin_op.prio * 10 + prio_inc
             }
-            _ => (&ops[*i1].bin_op.prio * 10, &ops[*i2].bin_op.prio * 10),
+            _ => &ops[bin_op_idx].bin_op.prio * 10,
         };
+    let mut indices: ExprIdxVec = (0..ops.len()).collect();
+    indices.sort_by(|i1, i2| {
+        let prio_i1 = prio_increase(*i1);
+        let prio_i2 = prio_increase(*i2);
         prio_i2.partial_cmp(&prio_i1).unwrap()
     });
     indices
 }
 
-fn prioritized_indices<T: Copy>(bin_ops: &[BinOp<T>], nodes: &Vec<Node<T>>) -> ExprIdxVec {
+fn prioritized_indices<T: Copy>(bin_ops: &[BinOp<T>], nodes: &[DeepNode<T>]) -> ExprIdxVec {
+    let prio_increase = |bin_op_idx: usize| match (&nodes[bin_op_idx], &nodes[bin_op_idx + 1]) {
+        (DeepNode::Num(_), DeepNode::Num(_)) => {
+            let prio_inc = 5;
+            &bin_ops[bin_op_idx].prio * 10 + prio_inc
+        }
+        _ => &bin_ops[bin_op_idx].prio * 10,
+    };
+
     let mut indices: ExprIdxVec = (0..bin_ops.len()).collect();
     indices.sort_by(|i1, i2| {
-        let (prio_i1, prio_i2) = match (&nodes[*i1], &nodes[*i2]) {
-            (Node::Num(_), Node::Num(_)) => {
-                let prio_inc = 5;
-                (
-                    &bin_ops[*i1].prio * 10 + prio_inc,
-                    &bin_ops[*i2].prio * 10 + prio_inc,
-                )
-            }
-            _ => (&bin_ops[*i1].prio * 10, &bin_ops[*i2].prio * 10),
-        };
+        let prio_i1 = prio_increase(*i1);
+        let prio_i2 = prio_increase(*i2);
         prio_i2.partial_cmp(&prio_i1).unwrap()
     });
     indices
 }
 
-impl<T: Copy + Debug> Expression<T> {
-    fn compile(&mut self) {
+impl<T: Copy + Debug> DeepEx<T> {
+    /// Evaluates all operators with numbers as operands.
+    pub fn compile(&mut self) {
+        // change from exression to number if an expression contains only a number
         for node in &mut self.nodes {
-            if let Node::Expr(ref mut e) = node {
-                e.compile();
+            if let DeepNode::Expr(ref e) = node {
+                if e.nodes.len() == 1 {
+                    match e.nodes[0] {
+                        DeepNode::Num(n) => {
+                            *node = DeepNode::Num(n);
+                        }
+                        _ => (),
+                    }
+                }
             };
         }
+        // after changing from expressions to numbers where possible the prios might change
+        self.prio_indices = prioritized_indices(&self.bin_ops, &self.nodes);
 
         let mut num_inds = self.prio_indices.clone();
         let mut used_prio_indices = ExprIdxVec::new();
@@ -333,8 +279,9 @@ impl<T: Copy + Debug> Expression<T> {
             let num_idx = num_inds[i];
             let node_1 = &self.nodes[num_idx];
             let node_2 = &self.nodes[num_idx + 1];
-            if let (Node::Num(num_1), Node::Num(num_2)) = (node_1, node_2) {
-                self.nodes[num_idx] = Node::Num((self.bin_ops[bin_op_idx].op)(*num_1, *num_2));
+            if let (DeepNode::Num(num_1), DeepNode::Num(num_2)) = (node_1, node_2) {
+                let bin_op_result = (self.bin_ops[bin_op_idx].apply)(*num_1, *num_2);
+                self.nodes[num_idx] = DeepNode::Num(bin_op_result);
                 self.nodes.remove(num_idx + 1);
                 // reduce indices after removed position
                 for num_idx_after in num_inds.iter_mut() {
@@ -356,23 +303,32 @@ impl<T: Copy + Debug> Expression<T> {
             .map(|x| *x.1)
             .collect();
 
+        if self.nodes.len() == 1 {
+            match self.nodes[0] {
+                DeepNode::Num(n) => {
+                    self.nodes[0] = DeepNode::Num(self.unary_op.apply(n));
+                    self.unary_op.clear();
+                }
+                _ => (),
+            }
+        }
         self.prio_indices = prioritized_indices(&self.bin_ops, &self.nodes);
     }
     pub fn new(
-        nodes: Vec<Node<T>>,
+        nodes: Vec<DeepNode<T>>,
         bin_ops: BinOpVec<T>,
-        unary_ops: CompositionOfUnaryOps<T>,
-    ) -> Result<Expression<T>, ExParseError> {
+        unary_op: UnaryOp<T>,
+    ) -> Result<DeepEx<T>, ExParseError> {
         if nodes.len() != bin_ops.len() + 1 {
             Err(ExParseError {
                 msg: "mismatch between number of nodes and binary operators".to_string(),
             })
         } else {
             let indices = prioritized_indices(&bin_ops, &nodes);
-            let mut expr = Expression {
+            let mut expr = DeepEx {
                 nodes: nodes,
                 bin_ops: bin_ops,
-                unary_ops: unary_ops,
+                unary_op,
                 prio_indices: indices,
             };
             expr.compile();
@@ -382,31 +338,90 @@ impl<T: Copy + Debug> Expression<T> {
     pub fn flatten(&self) -> FlatEx<T> {
         let (nodes, ops) = flatten_vecs(self, 0);
         let indices = prioritized_indices_flat(&ops, &nodes);
-        let n_unique_vars = nodes.iter().filter_map(|n| {
-            match n.kind {
-                FlatNodeKind::Var(idx) => Some(idx),
+        let mut found_vars = SmallVec::<[usize; 16]>::new();
+        let n_unique_vars = nodes
+            .iter()
+            .filter_map(|n| match n.kind {
+                FlatNodeKind::Var(idx) => {
+                    if !found_vars.contains(&idx) {
+                        found_vars.push(idx);
+                        Some(idx)
+                    } else {
+                        None
+                    }
+                }
                 _ => None,
-            }            
-        }).unique().count();
-        let mut flatex = FlatEx {
+            })
+            .count();
+        FlatEx {
             nodes: nodes,
             ops: ops,
             prio_indices: indices,
             n_unique_vars: n_unique_vars,
-        };
-        flatex.compile();
-        flatex
+        }
     }
 }
 
 #[cfg(test)]
-mod test {
+use crate::{make_default_operators, parse_with_default_ops, util::assert_float_eq_f64};
+#[test]
+fn test_flat_compile() {
+    let flat_ex = parse_with_default_ops::<f64>("1*sin(2-0.1)").unwrap();
+    assert_float_eq_f64(flat_ex.eval(&[]).unwrap(), 1.9f64.sin());
+    assert_eq!(flat_ex.nodes.len(), 1);
 
-    use crate::{parse_with_default_ops, util::assert_float_eq_f64};
-    #[test]
-    fn test_compile() {
-        let flat_ex = parse_with_default_ops::<f64>("1*sin(2-0.1)").unwrap();
-        assert_float_eq_f64(flat_ex.eval(&[]).unwrap(), 1.9f64.sin());
-        assert_eq!(flat_ex.nodes.len(), 1);
+    let flat_ex = parse_with_default_ops::<f64>("x*(2*(2*(2*4*8)))").unwrap();
+    assert_float_eq_f64(flat_ex.eval(&[1.0]).unwrap(), 2.0 * 2.0 * 2.0 * 4.0 * 8.0);
+    assert_eq!(flat_ex.nodes.len(), 2);
+
+    let flat_ex = parse_with_default_ops::<f64>("1*sin(2-0.1) + x").unwrap();
+    assert_float_eq_f64(flat_ex.eval(&[0.0]).unwrap(), 1.9f64.sin());
+    assert_eq!(flat_ex.nodes.len(), 2);
+    match flat_ex.nodes[0].kind {
+        FlatNodeKind::Num(n) => assert_float_eq_f64(n, 1.9f64.sin()),
+        _ => assert!(false),
+    }
+    match flat_ex.nodes[1].kind {
+        FlatNodeKind::Var(idx) => assert_eq!(idx, 0),
+        _ => assert!(false),
+    }
+
+    let flat_ex = parse_with_default_ops::<f64>("y + 1 - cos(1/(1*sin(2-0.1))-2) + 2 + x").unwrap();
+    assert_eq!(flat_ex.nodes.len(), 3);
+    match flat_ex.nodes[0].kind {
+        FlatNodeKind::Var(idx) => assert_eq!(idx, 0),
+        _ => assert!(false),
+    }
+    match flat_ex.nodes[1].kind {
+        FlatNodeKind::Num(_) => (),
+        _ => assert!(false),
+    }
+    match flat_ex.nodes[2].kind {
+        FlatNodeKind::Var(idx) => assert_eq!(idx, 1),
+        _ => assert!(false),
+    }
+}
+
+#[test]
+fn test_deep_compile() {
+    let ops = make_default_operators();
+    let nodes = vec![DeepNode::Num(4.5), DeepNode::Num(0.5), DeepNode::Num(1.4)];
+    let bin_ops: BinOpVec<f64> = smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()];
+    let unary_op = UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]);
+    let deep_ex = DeepEx::new(nodes, bin_ops, unary_op).unwrap();
+    let bin_ops: BinOpVec<f64> = smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()];
+    let unary_op = UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]);
+    let nodes = vec![
+        DeepNode::Num(4.5),
+        DeepNode::Num(0.5),
+        DeepNode::Expr(deep_ex),
+    ];
+    let deep_ex = DeepEx::new(nodes, bin_ops, unary_op).unwrap();
+    assert_eq!(deep_ex.nodes.len(), 1);
+    match deep_ex.nodes[0] {
+        DeepNode::Num(n) => assert_eq!(deep_ex.unary_op.apply(n), n),
+        _ => {
+            assert!(false);
+        }
     }
 }
