@@ -1,5 +1,5 @@
-#![doc(html_root_url = "https://docs.rs/exmex/0.7.1")]
-//! Exmex is a fast **ex**tendable **m**athematical **ex**pression evaluator.  
+#![doc(html_root_url = "https://docs.rs/exmex/0.8.1")]
+//! Exmex is a fast, simple, and **ex**tendable **m**athematical **ex**pression evaluator.  
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
@@ -17,7 +17,8 @@
 //! ## Variables
 //! For variables we can use strings that are not in the list of operators as shown in the following expression.
 //! Additionally, variables should consist only of letters, numbers, and underscores. More precisely, they need to fit the
-//! regular expression `r"^[a-zA-Z_]+[a-zA-Z_0-9]*"`.
+//! regular expression 
+//! ```r"^[a-zA-Z_]+[a-zA-Z_0-9]*"```.
 //! Variables' values are passed as slices to [`eval`](FlatEx::eval).
 //! ```rust
 //! # use std::error::Error;
@@ -33,8 +34,8 @@
 //! ```
 //! The `n`-th number in the slice corresponds to the `n`-th variable. Thereby only the
 //! first occurence of the variables is relevant. In this example, we have `z=2.5` and `y=3.7`.
-//! If variables are between curly brackets, they can have arbitrary names, e.g., 
-//! `{456/549*(}`, `{x}`, and `{x+y}`  are valid variable names as shown in the following.
+//! If variables are between curly brackets, they can have arbitrary names, e.g.,
+//! `{456/549*(}`, `{x}`, and confusingly even `{x+y}` are valid variable names as shown in the following.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
@@ -42,7 +43,7 @@
 //! use exmex::{make_default_operators, parse};
 //! let x = 2.1f64;
 //! let y = 0.1f64;
-//! let to_be_parsed = "log({x+y})";  // {x+y} is the name of one(!) variable, not the sum of two 😕.
+//! let to_be_parsed = "log({x+y})";  // {x+y} is the name of one(!) variable 😕.
 //! let expr = parse::<f64>(to_be_parsed, &make_default_operators::<f64>())?;
 //! assert!((expr.eval(&[x+y])? - 2.2f64.ln()).abs() < 1e-12);
 //! #
@@ -50,7 +51,7 @@
 //! # }
 //! ```
 //! ## Extendability
-//! Library users can also define a different set of operators as shown in the following.
+//! Library users can define their own set of operators as shown in the following.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
@@ -59,12 +60,12 @@
 //! let ops = [
 //!     Operator {
 //!         repr: "%",
-//!         bin_op: Some(BinOp{op: |a: i32, b: i32| a % b, prio: 1}),
+//!         bin_op: Some(BinOp{ apply: |a: i32, b: i32| a % b, prio: 1 }),
 //!         unary_op: None,
 //!     },
 //!     Operator {
 //!         repr: "/",
-//!         bin_op: Some(BinOp{op: |a: i32, b: i32| a / b, prio: 1}),
+//!         bin_op: Some(BinOp{ apply: |a: i32, b: i32| a / b, prio: 1 }),
 //!         unary_op: None,
 //!     },
 //! ];
@@ -84,9 +85,9 @@
 //! type [`Option<BinOp<T>>`](Operator::bin_op) and
 //! [`Option<fn(T) -> T>`](Operator::unary_op), respectively, as
 //! members. [`BinOp`](BinOp)
-//! contains in addition to the operator [`op`](BinOp::op) of type `fn(T, T) -> T` an
+//! contains in addition to the function pointer [`apply`](BinOp::apply) of type `fn(T, T) -> T` an
 //! integer [`prio`](BinOp::prio). Operators
-//! can be both, binary and unary such as `-` as defined in the list of default
+//! can be both, binary and unary. See, e.g.,  `-` defined in the list of default
 //! operators. Note that we expect a unary operator to be always on the left of a
 //! number.
 //!
@@ -106,12 +107,12 @@
 //! let ops = [
 //!     Operator {
 //!         repr: "&&",
-//!         bin_op: Some(BinOp{op: |a: bool, b: bool| a && b, prio: 1}),
+//!         bin_op: Some(BinOp{ apply: |a: bool, b: bool| a && b, prio: 1 }),
 //!         unary_op: None,
 //!     },
 //!     Operator {
 //!         repr: "||",
-//!         bin_op: Some(BinOp{op: |a: bool, b: bool| a || b, prio: 1}),
+//!         bin_op: Some(BinOp{ apply: |a: bool, b: bool| a || b, prio: 1 }),
 //!         unary_op: None,
 //!     },
 //!     Operator {
@@ -135,7 +136,7 @@
 //! If you want to be on the safe side, we suggest using parentheses.
 //!
 //! ## Unicode
-//! Unicode input strings are currently not supported 😕 but might be added in the 
+//! Unicode input strings are currently not supported 😕 but might be added in the
 //! future 😀.
 //!
 
@@ -191,7 +192,7 @@ mod tests {
                 Operator {
                     repr: "|",
                     bin_op: Some(BinOp {
-                        op: |a: u32, b: u32| a | b,
+                        apply: |a: u32, b: u32| a | b,
                         prio: 0,
                     }),
                     unary_op: None,
@@ -214,7 +215,8 @@ mod tests {
     fn test_variables_curly() {
         let operators = make_default_operators::<f64>();
 
-        let to_be_parsed = "5*{x} + 4*log2(log(1.5-{gamma}))*({x}*-(tan(cos(sin(652.2-{gamma}))))) + 3*{x}";
+        let to_be_parsed =
+            "5*{x} + 4*log2(log(1.5-{gamma}))*({x}*-(tan(cos(sin(652.2-{gamma}))))) + 3*{x}";
         let expr = parse(to_be_parsed, &operators).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 11.429314405093656);
         let to_be_parsed = "2*(4*{x} + y^2)";
@@ -232,7 +234,26 @@ mod tests {
     #[test]
     fn test_variables() {
         let operators = make_default_operators::<f64>();
-        
+
+        let to_be_parsed = "sin(sin(x - 1 / sin(y * 5)) + (5.0 - 1/z))";
+        let expr = parse(to_be_parsed, &operators).unwrap();
+        let reference = |x: f64, y: f64, z: f64| {
+             ((x - 1.0 / (y * 5.0).sin()).sin() + (5.0 - 1.0 / z)).sin()
+        };
+        assert_float_eq_f64(
+            expr.eval(&[1.0, 2.0, 4.0]).unwrap(),
+            reference(1.0, 2.0, 4.0),
+        );
+
+        let to_be_parsed = "0.02*sin(-(3*(2*(5.0 - 1/z))))";
+        let expr = parse(to_be_parsed, &operators).unwrap();
+        let reference = |z: f64| 0.02 * (-(3.0 * (2.0 * (5.0 - 1.0 / z)))).sin();
+        assert_float_eq_f64(expr.eval(&[4.0]).unwrap(), reference(4.0));
+
+        let to_be_parsed = "y + 1 + 0.5 * x";
+        let expr = parse(to_be_parsed, &operators).unwrap();
+        assert_float_eq_f64(expr.eval(&[1.0, 3.0]).unwrap(), 3.5);
+
         let to_be_parsed = " -(-(1+x))";
         let expr = parse(to_be_parsed, &operators).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 2.0);
@@ -245,7 +266,6 @@ mod tests {
         let expr = parse(to_be_parsed, &operators).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5, 0.2532]).unwrap(), -3.1164569260604176);
 
-
         let to_be_parsed = "5*x + 4*y + 3*x";
         let expr = parse(to_be_parsed, &operators).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 8.0);
@@ -305,7 +325,6 @@ mod tests {
         let to_be_parsed = "trunc(x) + fract(x)";
         let expr = parse(to_be_parsed, &operators).unwrap();
         assert_float_eq_f64(expr.eval(&[23422.52345]).unwrap(), 23422.52345);
-
     }
 
     #[test]
@@ -332,7 +351,7 @@ mod tests {
             Operator {
                 repr: "**",
                 bin_op: Some(BinOp {
-                    op: |a: f32, b| a.powf(b),
+                    apply: |a: f32, b| a.powf(b),
                     prio: 2,
                 }),
                 unary_op: None,
@@ -340,7 +359,7 @@ mod tests {
             Operator {
                 repr: "*",
                 bin_op: Some(BinOp {
-                    op: |a, b| a * b,
+                    apply: |a, b| a * b,
                     prio: 1,
                 }),
                 unary_op: None,
@@ -358,7 +377,7 @@ mod tests {
         let zero_mapper = Operator {
             repr: "zer0",
             bin_op: Some(BinOp {
-                op: |_: f32, _| 0.0,
+                apply: |_: f32, _| 0.0,
                 prio: 2,
             }),
             unary_op: Some(|_| 0.0),
