@@ -1,4 +1,5 @@
 use num::Float;
+use smallvec::{SmallVec, smallvec};
 
 /// Operators can be custom-defined by the library-user in terms of this struct.
 ///
@@ -10,7 +11,7 @@ use num::Float;
 ///     Operator {
 ///         repr: "-",
 ///         bin_op: Some(BinOp {
-///             op: |a, b| a - b,
+///             apply: |a, b| a - b,
 ///             prio: 0,
 ///         }),
 ///         unary_op: Some(|a: f32| (-a)),
@@ -34,11 +35,65 @@ pub struct Operator<'a, T> {
     pub unary_op: Option<fn(T) -> T>,
 }
 
+
+pub type VecOfUnaryFuncs<T> = SmallVec<[fn(T) -> T; 8]>;
+
+/// Container of unary operators of one expression
+#[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
+pub struct UnaryOp<T> {
+    funcs_to_be_composed: VecOfUnaryFuncs<T>,
+}
+
+impl<T> UnaryOp<T> {
+    /// Applies unary operators one after the other starting with the last.
+    /// # Arguments
+    ///
+    /// * `x` - number the unary operators are applied to
+    ///
+    pub fn apply(&self, x: T) -> T {
+        let mut result = x;
+        // rev, since the last uop is applied first by convention
+        for uo in self.funcs_to_be_composed.iter().rev() {
+            result = uo(result);
+        }
+        result
+    }
+
+    pub fn append_front(&mut self, other: &mut UnaryOp<T>) {
+        self.funcs_to_be_composed = other
+            .funcs_to_be_composed
+            .iter()
+            .chain(self.funcs_to_be_composed.iter())
+            .map(|f| *f)
+            .collect::<SmallVec<_>>();
+    }
+
+    pub fn len(&self) -> usize {
+        self.funcs_to_be_composed.len()
+    }
+
+    pub fn new() -> Self {
+        Self {
+            funcs_to_be_composed: smallvec![],
+        }
+    }
+
+    pub fn from_vec(v: VecOfUnaryFuncs<T>) -> Self {
+        Self {
+            funcs_to_be_composed: v,
+        }
+    }
+
+    pub fn clear(&mut self) {
+        self.funcs_to_be_composed.clear();
+    }
+}
+
 /// A binary operator that consists of a function pointer and a priority.
 #[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub struct BinOp<T> {
     /// Implementation of the binary operation, e.g., `|a, b| a * b` for multiplication.
-    pub op: fn(T, T) -> T,
+    pub apply: fn(T, T) -> T,
     /// Priority of the binary operation. A binary operation with a
     /// higher number will be executed first. For instance, in a sane world `*`
     /// has a higher priority than `+`. However, in Exmex land you could also define
@@ -52,7 +107,7 @@ pub fn make_default_operators<'a, T: Float>() -> [Operator<'a, T>; 23] {
         Operator {
             repr: "^",
             bin_op: Some(BinOp {
-                op: |a: T, b| a.powf(b),
+                apply: |a: T, b| a.powf(b),
                 prio: 2,
             }),
             unary_op: None,
@@ -60,7 +115,7 @@ pub fn make_default_operators<'a, T: Float>() -> [Operator<'a, T>; 23] {
         Operator {
             repr: "*",
             bin_op: Some(BinOp {
-                op: |a, b| a * b,
+                apply: |a, b| a * b,
                 prio: 1,
             }),
             unary_op: None,
@@ -68,7 +123,7 @@ pub fn make_default_operators<'a, T: Float>() -> [Operator<'a, T>; 23] {
         Operator {
             repr: "/",
             bin_op: Some(BinOp {
-                op: |a, b| a / b,
+                apply: |a, b| a / b,
                 prio: 1,
             }),
             unary_op: None,
@@ -76,7 +131,7 @@ pub fn make_default_operators<'a, T: Float>() -> [Operator<'a, T>; 23] {
         Operator {
             repr: "+",
             bin_op: Some(BinOp {
-                op: |a, b| a + b,
+                apply: |a, b| a + b,
                 prio: 0,
             }),
             unary_op: Some(|a: T| a),
@@ -84,7 +139,7 @@ pub fn make_default_operators<'a, T: Float>() -> [Operator<'a, T>; 23] {
         Operator {
             repr: "-",
             bin_op: Some(BinOp {
-                op: |a, b| a - b,
+                apply: |a, b| a - b,
                 prio: 0,
             }),
             unary_op: Some(|a: T| (-a)),
