@@ -1,7 +1,6 @@
-use crate::expression::{BinOpVec, Expression, FlatEx, Node, N_NODES_ON_STACK};
-use crate::operators::{make_default_operators, BinOp, Operator};
-use crate::util::{apply_unary_ops, CompositionOfUnaryOps};
-use itertools::Itertools;
+use crate::expression::{BinOpVec, DeepEx, DeepNode, FlatEx, N_NODES_ON_STACK};
+use crate::operators::{make_default_operators, BinOp, Operator, UnaryOp, VecOfUnaryFuncs};
+// use itertools::Itertools;
 use lazy_static::lazy_static;
 use num::Float;
 use regex::Regex;
@@ -77,7 +76,7 @@ fn is_numeric_regex<'a>(re: &Regex, text: &'a str) -> Option<&'a str> {
 ///
 /// See [`parse_with_number_pattern`](parse_with_number_pattern)
 ///
-fn apply_regexes<'a, 'b, T: Copy + FromStr + Debug, F: Fn(&'b str) -> Option<&'b str>>(
+fn parsed_tokens<'a, 'b, T: Copy + FromStr + Debug, F: Fn(&'b str) -> Option<&'b str>>(
     text: &'b str,
     ops_in: &[Operator<'a, T>],
     is_numeric: F,
@@ -85,7 +84,6 @@ fn apply_regexes<'a, 'b, T: Copy + FromStr + Debug, F: Fn(&'b str) -> Option<&'b
 where
     <T as std::str::FromStr>::Err: Debug,
 {
-       
     // We sort operators inverse alphabetically such that log2 has higher priority than log (wlog :D).
 
     let mut ops_tmp = ops_in.iter().clone().collect::<SmallVec<[_; 64]>>();
@@ -95,7 +93,7 @@ where
     lazy_static! {
         static ref RE_NAME: Regex = Regex::new(r"^[a-zA-Z_]+[a-zA-Z_0-9]*").unwrap();
     }
-    
+
     let mut cur_offset = 0usize;
     let find_ops = |offset: usize| {
         ops.iter().find(|op| {
@@ -178,9 +176,9 @@ where
 ///
 fn make_expression<T>(
     parsed_tokens: &[ParsedToken<T>],
-    parsed_vars: &[String],
-    unary_ops: CompositionOfUnaryOps<T>,
-) -> Result<(Expression<T>, usize), ExParseError>
+    parsed_vars: &[&String],
+    unary_ops: UnaryOp<T>,
+) -> Result<(DeepEx<T>, usize), ExParseError>
 where
     T: Copy + FromStr + Debug,
 {
@@ -210,7 +208,7 @@ where
     // variable 'tokens' from the outer scope
     let process_unary = |i: usize, uo| {
         // gather subsequent unary operators from the beginning
-        let uops = once(uo)
+        let vec_of_uops = once(uo)
             .chain(
                 (i + 1..parsed_tokens.len())
                     .map(|j| match parsed_tokens[j] {
@@ -220,9 +218,9 @@ where
                     .take_while(|uo_| uo_.is_some())
                     .flatten(),
             )
-            .collect::<CompositionOfUnaryOps<_>>();
-        let n_uops = uops.len();
-
+            .collect::<VecOfUnaryFuncs<_>>();
+        let n_uops = vec_of_uops.len();
+        let uop = UnaryOp::from_vec(vec_of_uops);
         match &parsed_tokens[i + n_uops] {
             ParsedToken::Paren(p) => match p {
                 Paren::Close => Err(ExParseError {
@@ -230,19 +228,19 @@ where
                 }),
                 Paren::Open => {
                     let (expr, i_forward) =
-                        make_expression::<T>(&parsed_tokens[i + n_uops + 1..], &parsed_vars, uops)?;
-                    Ok((Node::Expr(expr), i_forward + n_uops + 1))
+                        make_expression::<T>(&parsed_tokens[i + n_uops + 1..], &parsed_vars, uop)?;
+                    Ok((DeepNode::Expr(expr), i_forward + n_uops + 1))
                 }
             },
             ParsedToken::Var(name) => {
-                let expr = Expression::new(
-                    vec![Node::Var(find_var_index(&name))],
+                let expr = DeepEx::new(
+                    vec![DeepNode::Var(find_var_index(&name))],
                     BinOpVec::new(),
-                    uops,
+                    uop,
                 )?;
-                Ok((Node::Expr(expr), n_uops + 1))
+                Ok((DeepNode::Expr(expr), n_uops + 1))
             }
-            ParsedToken::Num(n) => Ok((Node::Num(apply_unary_ops(&uops, *n)), n_uops + 1)),
+            ParsedToken::Num(n) => Ok((DeepNode::Num(uop.apply(*n)), n_uops + 1)),
             ParsedToken::Op(_) => Err(ExParseError {
                 msg: "a unary operator cannot be followed by a binary operator".to_string(),
             }),
@@ -250,7 +248,7 @@ where
     };
 
     let mut bin_ops = BinOpVec::new();
-    let mut nodes = Vec::<Node<T>>::new();
+    let mut nodes = Vec::<DeepNode<T>>::new();
 
     // The main loop checks one token after the next whereby sub-expressions are
     // handled recursively. Thereby, the token-position-index idx_tkn is increased
@@ -298,11 +296,11 @@ where
                 }
             },
             ParsedToken::Num(n) => {
-                nodes.push(Node::Num(*n));
+                nodes.push(DeepNode::Num(*n));
                 idx_tkn += 1;
             }
             ParsedToken::Var(name) => {
-                nodes.push(Node::Var(find_var_index(&name)));
+                nodes.push(DeepNode::Var(find_var_index(&name)));
                 idx_tkn += 1;
             }
             ParsedToken::Paren(p) => match p {
@@ -311,9 +309,9 @@ where
                     let (expr, i_forward) = make_expression::<T>(
                         &parsed_tokens[idx_tkn..],
                         &parsed_vars,
-                        CompositionOfUnaryOps::new(),
+                        UnaryOp::new(),
                     )?;
-                    nodes.push(Node::Expr(expr));
+                    nodes.push(DeepNode::Expr(expr));
                     idx_tkn += i_forward;
                 }
                 Paren::Close => {
@@ -323,7 +321,7 @@ where
             },
         }
     }
-    Ok((Expression::new(nodes, bin_ops, unary_ops)?, idx_tkn))
+    Ok((DeepEx::new(nodes, bin_ops, unary_ops)?, idx_tkn))
 }
 
 /// Tries to give useful error messages for invalid constellations of the parsed tokens
@@ -444,22 +442,25 @@ where
 fn parsed_tokens_to_flatex<T: Copy + FromStr + Debug>(
     parsed_tokens: &SmallVec<[ParsedToken<T>; 2 * N_NODES_ON_STACK]>,
 ) -> Result<FlatEx<T>, ExParseError> {
+    let mut found_vars = SmallVec::<[&str; 16]>::new();
     let parsed_vars = parsed_tokens
         .iter()
         .filter_map(|pt| match pt {
-            ParsedToken::Var(name) => Some(name.clone()),
+            ParsedToken::Var(name) => {
+                if !found_vars.contains(&name.as_str()) {
+                    found_vars.push(name.as_str());
+                    Some(name)
+                } else {
+                    None
+                }
+            }
             _ => None,
         })
-        .unique()
         .collect::<SmallVec<[_; N_NODES_ON_STACK]>>();
 
     check_preconditions(&parsed_tokens[..])?;
 
-    let (expr, _) = make_expression(
-        &parsed_tokens[0..],
-        &parsed_vars,
-        CompositionOfUnaryOps::new(),
-    )?;
+    let (expr, _) = make_expression(&parsed_tokens[0..], &parsed_vars, UnaryOp::new())?;
     Ok(expr.flatten())
 }
 
@@ -474,7 +475,7 @@ where
     <T as std::str::FromStr>::Err: Debug,
     T: Copy + FromStr + Debug,
 {
-    let parsed_tokens = apply_regexes(text, ops, is_numeric_text)?;
+    let parsed_tokens = parsed_tokens(text, ops, is_numeric_text)?;
     parsed_tokens_to_flatex(&parsed_tokens)
 }
 
@@ -509,7 +510,7 @@ where
 /// * in `parsed_tokens` a closing parentheses is directly following an operator, e.g., `+)`, or
 /// * a unary operator is followed directly by a binary operator, e.g., `sin*`.
 ///
-pub fn parse_with_number_pattern<'a,'b, T>(
+pub fn parse_with_number_pattern<'a, 'b, T>(
     text: &'b str,
     ops: &[Operator<'a, T>],
     number_regex_pattern: &str,
@@ -528,7 +529,7 @@ where
         }
     };
     let is_numeric = |text: &'b str| is_numeric_regex(&re_number, &text);
-    let parsed_tokens = apply_regexes(text, ops, is_numeric)?;
+    let parsed_tokens = parsed_tokens(text, ops, is_numeric)?;
     parsed_tokens_to_flatex(&parsed_tokens)
 }
 
@@ -549,13 +550,16 @@ where
 
 #[cfg(test)]
 mod tests {
-    use crate::{ExParseError, parse::{apply_regexes, check_preconditions, is_numeric_text, make_default_operators}};
+    use crate::{
+        parse::{check_preconditions, is_numeric_text, make_default_operators, parsed_tokens},
+        ExParseError,
+    };
 
     #[test]
     fn test_apply_regexes() {
         let text = r"5\6";
         let ops = make_default_operators::<f32>();
-        let elts = apply_regexes(text, &ops, is_numeric_text);
+        let elts = parsed_tokens(text, &ops, is_numeric_text);
         assert!(elts.is_err());
     }
 
@@ -572,7 +576,7 @@ mod tests {
                 }
             }
             let ops = make_default_operators::<f32>();
-            let elts = apply_regexes(text, &ops, is_numeric_text);
+            let elts = parsed_tokens(text, &ops, is_numeric_text);
             match elts {
                 Ok(elts_unwr) => {
                     let err = check_preconditions(&elts_unwr[..]);
